//
//  OCNewsHelper.m
//  iOCNews
//

/************************************************************************
 
 Copyright 2013 Peter Hedlund peter.hedlund@me.com
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
 notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
 notice, this list of conditions and the following disclaimer in the
 documentation and/or other materials provided with the distribution.
 
 THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 *************************************************************************/

#import "OCNewsHelper.h"
#import "OCAPIClient.h"
#import "Feeds.h"
#import "NSDictionary+HandleNull.h"

// the type of the query (Feed: 0, Folder: 1, Starred: 2, All: 3)
const int UPDATE_FEED = 0;
const int UPDATE_FOLDER = 1;
const int UPDATE_STARRED = 2;
const int UPDATE_ALL = 3;

@interface OCNewsHelper () {
    NSMutableArray *foldersToAdd;
    NSMutableArray *foldersToDelete;
    NSMutableArray *foldersToRename;
    NSMutableArray *feedsToAdd;
    NSMutableArray *feedsToDelete;
    NSMutableArray *feedsToMove;
    NSMutableArray *itemsToMarkRead;
    NSMutableArray *itemsToMarkUnread;
    NSMutableArray *itemsToStar;
    NSMutableArray *itemsToUnstar;
}

- (int)addFolderFromDictionary:(NSDictionary*)dict;
- (int)addFeedFromDictionary:(NSDictionary*)dict;
- (void)addItemFromDictionary:(NSDictionary*)dict;
- (NSNumber*)folderLastModified:(NSNumber*)folderId;
- (NSNumber*)feedLastModified:(NSNumber*)feedId;
- (void)updateItemsFirstTime;
- (void)updateItemsWithLastModified:(NSNumber*)lastMod type:(NSNumber*)aType andId:(NSNumber*)anId;
- (void)updateFeedItemsWithLastModified:(NSNumber*)lastMod type:(NSNumber*)aType andId:(NSNumber*)anId;

@end

@implementation OCNewsHelper

@synthesize context;
@synthesize objectModel;
@synthesize coordinator;
@synthesize feedsRequest;
@synthesize folderRequest;
@synthesize feedRequest;
@synthesize itemRequest;

+ (OCNewsHelper*)sharedHelper {
    static dispatch_once_t once_token;
    static id sharedHelper;
    dispatch_once(&once_token, ^{
        sharedHelper = [[OCNewsHelper alloc] init];
    });
    return sharedHelper;
}

- (OCNewsHelper*)init {
    self = [super init];
    if (!self) {
        return nil;
    }
    
    NSError *error = nil;
    NSArray *feeds = [self.context executeFetchRequest:self.feedsRequest error:&error];
    if (!feeds.count) {
        Feeds *newFeeds = [NSEntityDescription insertNewObjectForEntityForName:@"Feeds" inManagedObjectContext:self.context];
        newFeeds.starredCount = [NSNumber numberWithInt:0];
        newFeeds.newestItemId = [NSNumber numberWithInt:0];
    }
    
    error = nil;
    NSArray *feed = [self.context executeFetchRequest:self.feedRequest error:&error];
    
    if (!feed.count) {
        Feed *allFeed = [NSEntityDescription insertNewObjectForEntityForName:@"Feed" inManagedObjectContext:self.context];
        allFeed.myId = [NSNumber numberWithInt:-2];
        allFeed.url = @"";
        allFeed.title = @"All Articles";
        allFeed.faviconLink = @"favicon";
        allFeed.added = [NSNumber numberWithInt:1];
        allFeed.folderId = [NSNumber numberWithInt:0];
        allFeed.unreadCount = [NSNumber numberWithInt:0];
        allFeed.link = @"";
        
        Feed *starredFeed = [NSEntityDescription insertNewObjectForEntityForName:@"Feed" inManagedObjectContext:self.context];
        starredFeed.myId = [NSNumber numberWithInt:-1];
        starredFeed.url = @"";
        starredFeed.title = @"Starred";
        starredFeed.faviconLink = @"star_icon";
        starredFeed.added = [NSNumber numberWithInt:2];
        starredFeed.folderId = [NSNumber numberWithInt:0];
        starredFeed.unreadCount = [NSNumber numberWithInt:0];
        starredFeed.link = @"";
        [self addFeedExtra:starredFeed];
        starredFeed.extra.lastModified = [NSNumber numberWithInt:[[NSUserDefaults standardUserDefaults] integerForKey:@"LastModified"]];
    }
    
    NSUserDefaults *prefs = [NSUserDefaults standardUserDefaults];
    foldersToAdd = [[prefs arrayForKey:@"FoldersToAdd"] mutableCopy];
    foldersToDelete = [[prefs arrayForKey:@"FoldersToDelete"] mutableCopy];
    foldersToRename = [[prefs arrayForKey:@"FoldersToRename"] mutableCopy];
    feedsToAdd = [[prefs arrayForKey:@"FeedsToAdd"] mutableCopy];
    feedsToDelete = [[prefs arrayForKey:@"FeedsToDelete"] mutableCopy];
    feedsToMove = [[prefs arrayForKey:@"FeedsToMove"] mutableCopy];
    itemsToMarkRead = [[prefs arrayForKey:@"ItemsToMarkRead"] mutableCopy];
    itemsToMarkUnread = [[prefs arrayForKey:@"ItemsToMarkUnread"] mutableCopy];
    itemsToStar = [[prefs arrayForKey:@"ItemsToStar"] mutableCopy];
    itemsToUnstar = [[prefs arrayForKey:@"ItemsToUnstar"] mutableCopy];

    [self updateStarredCount];
    [self saveContext];

    __unused int status = [[OCAPIClient sharedClient] networkReachabilityStatus];

    return self;
}

- (NSManagedObjectModel *)objectModel {
    if (!objectModel) {
        NSURL *modelURL = [[NSBundle mainBundle] URLForResource:@"News" withExtension:@"momd"];
        objectModel = [[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL];
    }
    return objectModel;
}

- (NSPersistentStoreCoordinator *)coordinator {
    if (!coordinator) {
        NSURL *storeURL = [self documentsDirectoryURL];
        storeURL = [storeURL URLByAppendingPathComponent:@"News.sqlite"];
        NSDictionary *options = @{NSMigratePersistentStoresAutomaticallyOption : @YES,
                                  NSInferMappingModelAutomaticallyOption : @YES };
        NSError *error = nil;
        coordinator = [[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:[self objectModel]];
        if (![coordinator addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:storeURL options:options error:&error]) {
            NSLog(@"Error %@, %@", error, [error localizedDescription]);
            abort();
        }
    }
    return coordinator;
}

- (NSManagedObjectContext *)context {
    if (!context) {
        NSPersistentStoreCoordinator *myCoordinator = [self coordinator];
        if (myCoordinator != nil) {
            context = [[NSManagedObjectContext alloc] init];
            [context setPersistentStoreCoordinator:myCoordinator];
        }
    }
    return context;
}

- (void)saveContext {
    NSUserDefaults *prefs = [NSUserDefaults standardUserDefaults];
    [prefs setObject:foldersToAdd forKey:@"FoldersToAdd"];
    [prefs setObject:foldersToDelete forKey:@"FoldersToDelete"];
    [prefs setObject:foldersToRename forKey:@"FoldersToRename"];
    [prefs setObject:feedsToAdd forKey:@"FeedsToAdd"];
    [prefs setObject:feedsToDelete forKey:@"FeedsToDelete"];
    [prefs setObject:feedsToMove forKey:@"FeedsToMove"];
    [prefs setObject:itemsToMarkRead forKey:@"ItemsToMarkRead"];
    [prefs setObject:itemsToMarkUnread forKey:@"ItemsToMarkUnread"];
    [prefs setObject:itemsToStar forKey:@"ItemsToStar"];
    [prefs setObject:itemsToUnstar forKey:@"ItemsToUnstar"];
    [prefs synchronize];
    
    NSError *error = nil;
    if (self.context != nil) {
        if ([self.context hasChanges] && ![self.context save:&error]) {
            NSLog(@"Error saving data %@, %@", error, [error userInfo]);
            //abort();
        } else {
            NSLog(@"Data saved");
        }
    }
}

#pragma mark - COREDATA -INSERT

- (int)addFolderFromDictionary:(NSDictionary*)dict {
    Folder *newFolder = [NSEntityDescription insertNewObjectForEntityForName:@"Folder" inManagedObjectContext:self.context];
    newFolder.myId = [dict objectForKey:@"id"];
    newFolder.name = [dict objectForKeyNotNull:@"name" fallback:@"Folder"];
    newFolder.unreadCountValue = 0;
    return newFolder.myIdValue;
}

- (int)addFeedFromDictionary:(NSDictionary *)dict {
    Feed *newFeed = [NSEntityDescription insertNewObjectForEntityForName:@"Feed" inManagedObjectContext:self.context];
    newFeed.myId = [dict objectForKey:@"id"];
    newFeed.url = [dict objectForKeyNotNull:@"url" fallback:@""];
    newFeed.title = [dict objectForKeyNotNull:@"title" fallback:@""];
    newFeed.faviconLink = [dict objectForKeyNotNull:@"faviconLink" fallback:@"favicon"];
    newFeed.added = [dict objectForKey:@"added"];
    newFeed.folderId = [dict objectForKey:@"folderId"];
    newFeed.unreadCount = [dict objectForKey:@"unreadCount"];
    newFeed.link = [dict objectForKeyNotNull:@"link" fallback:@""];
    [self addFeedExtra:newFeed];
    return newFeed.myIdValue;
}

- (void)addItemFromDictionary:(NSDictionary *)dict {
    Item *newItem = [NSEntityDescription insertNewObjectForEntityForName:@"Item" inManagedObjectContext:self.context];
    newItem.myId = [dict objectForKey:@"id"];
    newItem.guid = [dict objectForKey:@"guid"];
    newItem.guidHash = [dict objectForKey:@"guidHash"];
    newItem.url = [dict objectForKeyNotNull:@"url" fallback:@""];
    newItem.title = [dict objectForKeyNotNull:@"title" fallback:@""];
    newItem.author = [dict objectForKeyNotNull:@"author" fallback:@""];
    newItem.pubDate = [dict objectForKeyNotNull:@"pubDate" fallback:nil];
    newItem.body = [dict objectForKeyNotNull:@"body" fallback:@""];
    newItem.enclosureMime = [dict objectForKeyNotNull:@"enclosureMime" fallback:@""];
    newItem.enclosureLink = [dict objectForKeyNotNull:@"enclosureLink" fallback:@""];
    newItem.feedId = [dict objectForKey:@"feedId"];
    newItem.unread = [dict objectForKey:@"unread"];
    newItem.starred = [dict objectForKey:@"starred"];
    newItem.lastModified = [dict objectForKey:@"lastModified"];
    [self addItemExtra:newItem];
}

- (int)addFolder:(id)JSON {
    NSDictionary *jsonDict = (NSDictionary *) JSON;
    NSMutableArray *newFolders = [jsonDict objectForKey:@"folders"];
    int newFolderId = [self addFolderFromDictionary:[newFolders lastObject]];
    [self updateTotalUnreadCount];
    return newFolderId;
}

- (void)deleteFolder:(Folder*)folder {
    if (folder) {
        self.feedRequest.predicate = [NSPredicate predicateWithFormat:@"folderId == %@", folder.myId];
        NSMutableArray *feedsToBeDeleted = [NSMutableArray arrayWithArray:[self.context executeFetchRequest:self.feedRequest error:nil]];
        while (feedsToBeDeleted.count > 0) {
            [self deleteFeed:[feedsToDelete lastObject]];
            [feedsToBeDeleted removeLastObject];
        }
        [self.context deleteObject:folder];
        [self updateTotalUnreadCount];
    }
}

- (NSArray*)folders {
    self.folderRequest.predicate = nil;
    return [self.context executeFetchRequest:self.folderRequest error:nil];
}

- (int)addFeed:(id)JSON {
    NSDictionary *jsonDict = (NSDictionary *) JSON;
    NSMutableArray *newFeeds = [jsonDict objectForKey:@"feeds"];
    int newFeedId = [self addFeedFromDictionary:[newFeeds lastObject]];
    [self updateTotalUnreadCount];
    return newFeedId;
}

- (void)addFeedExtra:(Feed *)feed {
    FeedExtra *extra = [NSEntityDescription insertNewObjectForEntityForName:@"FeedExtra" inManagedObjectContext:self.context];
    extra.displayTitle = feed.title;
    extra.parent = feed;
    feed.extra = extra;
}

- (void)addItemExtra:(Item *)item {
    ItemExtra *extra = [NSEntityDescription insertNewObjectForEntityForName:@"ItemExtra" inManagedObjectContext:self.context];
    extra.parent = item;
    item.extra = extra;
}

- (void)deleteFeed:(Feed*)feed {
    if (feed) {
        [self.itemRequest setPredicate:[NSPredicate predicateWithFormat:@"feedId == %@", feed.myId]];
        
        NSError *error = nil;
        NSArray *feedItems = [self.context executeFetchRequest:self.itemRequest error:&error];
        for (Item *item in feedItems) {
            [self.context deleteObject:item];
        }
        [self.context deleteObject:feed];
        [self updateTotalUnreadCount];
    }
}

- (void)sync {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        [[OCAPIClient sharedClient] getPath:@"feeds" parameters:nil success:^(AFHTTPRequestOperation *operation, id responseObject) {
            [self updateFeeds:responseObject];
            [self updateFolders];
        } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
            NSString *message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Updating Feeds", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        }];
    } else {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Unable to Reach Server", @"Title",
                                  @"Please check network connection and login.", @"Message", nil];
        [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
    }
}

- (void)updateFolders {
    [[OCAPIClient sharedClient] getPath:@"folders" parameters:nil success:^(AFHTTPRequestOperation *operation, id responseObject) {
        //Remove previous
        //TODO: only fetch myId
        NSError *error = nil;
        [self.folderRequest setPredicate:nil];
        NSArray *oldFolders = [self.context executeFetchRequest:self.folderRequest error:&error];
        NSArray *knownIds = [oldFolders valueForKey:@"myId"];
        
        NSLog(@"Count: %i", oldFolders.count);
        
        //Add the new folders
        NSDictionary *folderDict = (NSDictionary *)responseObject;
        
        NSArray *newFolders = [NSArray arrayWithArray:[folderDict objectForKey:@"folders"]];
        
        NSArray *newIds = [newFolders valueForKey:@"id"];
        NSLog(@"Known: %@; New: %@", knownIds, newIds);
        
        NSMutableArray *newOnServer = [NSMutableArray arrayWithArray:newIds];
        [newOnServer removeObjectsInArray:knownIds];
        NSLog(@"New on server: %@", newOnServer);
        [newOnServer enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
            NSPredicate * predicate = [NSPredicate predicateWithFormat:@"id == %@", obj];
            NSArray * matches = [newFolders filteredArrayUsingPredicate:predicate];
            if (matches.count > 0) {
                [self addFolderFromDictionary:[matches lastObject]];
            }
        }];
        
        NSMutableArray *deletedOnServer = [NSMutableArray arrayWithArray:knownIds];
        [deletedOnServer removeObjectsInArray:newIds];
        [deletedOnServer filterUsingPredicate:[NSPredicate predicateWithFormat:@"self >= 0"]];
        NSLog(@"Deleted on server: %@", deletedOnServer);
        while (deletedOnServer.count > 0) {
            Folder *folderToRemove = [self folderWithId:[deletedOnServer lastObject]];
            [self.context deleteObject:folderToRemove];
            [deletedOnServer removeLastObject];
        }

        for (NSNumber *folderId in foldersToDelete) {
            Folder *folder = [self folderWithId:folderId];
            [self deleteFolderOffline:folder]; //the feed will have been readded as new on server
        }
        [foldersToDelete removeAllObjects];
        
        for (NSString *name in foldersToAdd) {
            [self addFolderOffline:name];
        }
        [foldersToAdd removeAllObjects];
        
        //@{@"folderId": anId, @"name": newName}
        for (NSDictionary *dict in foldersToRename) {
            [self renameFolderOfflineWithId:[dict objectForKey:@"folderId"] To:[dict objectForKey:@"name"]];
        }
        [foldersToRename removeAllObjects];
        NSNumber *lastMod = [NSNumber numberWithInt:[[NSUserDefaults standardUserDefaults] integerForKey:@"LastModified"]];
        if ([self itemCount] > 0) {
            [self updateItemsWithLastModified:lastMod type:[NSNumber numberWithInt:UPDATE_ALL] andId:[NSNumber numberWithInt:0]];
        } else {
            [self updateItemsFirstTime];
            [self updateItemsWithLastModified:lastMod type:[NSNumber numberWithInt:UPDATE_STARRED] andId:[NSNumber numberWithInt:0]];
        }
        [self updateTotalUnreadCount];
    
    } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
        NSString *message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Updating Feeds", @"Title", message, @"Message", nil];
        [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
    }];
}

- (void)updateFeeds:(id)JSON {
    //Remove previous
    //TODO: only fetch myId
    NSError *error = nil;
    [self.feedRequest setPredicate:nil];
    NSArray *oldFeeds = [self.context executeFetchRequest:self.feedRequest error:&error];
    NSArray *knownIds = [oldFeeds valueForKey:@"myId"];
    
    NSLog(@"Count: %i", oldFeeds.count);
    
    error = nil;
    NSArray *feeds = [self.context executeFetchRequest:self.feedsRequest error:&error];
    
    //Add the new feeds
    NSDictionary *jsonDict = (NSDictionary *) JSON;
    Feeds *theFeeds = [feeds objectAtIndex:0];
    theFeeds.starredCount = [jsonDict objectForKey:@"starredCount"];
    theFeeds.newestItemId = [jsonDict objectForKey:@"newestItemId"];
    
    NSArray *newFeeds = [NSArray arrayWithArray:[jsonDict objectForKey:@"feeds"]];
    
    NSArray *newIds = [newFeeds valueForKey:@"id"];
    NSLog(@"Known: %@; New: %@", knownIds, newIds);
    
    NSMutableArray *newOnServer = [NSMutableArray arrayWithArray:newIds];
    [newOnServer removeObjectsInArray:knownIds];
    NSLog(@"New on server: %@", newOnServer);
    [newOnServer enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
        NSPredicate * predicate = [NSPredicate predicateWithFormat:@"id == %@", obj];
        NSArray * matches = [newFeeds filteredArrayUsingPredicate:predicate];
        if (matches.count > 0) {
            [self addFeedFromDictionary:[matches lastObject]];
        }
    }];
    
    NSMutableArray *deletedOnServer = [NSMutableArray arrayWithArray:knownIds];
    [deletedOnServer removeObjectsInArray:newIds];
    [deletedOnServer filterUsingPredicate:[NSPredicate predicateWithFormat:@"self >= 0"]];
    NSLog(@"Deleted on server: %@", deletedOnServer);
    while (deletedOnServer.count > 0) {
        Feed *feedToRemove = [self feedWithId:[deletedOnServer lastObject]];
        [self.context deleteObject:feedToRemove];
        [deletedOnServer removeLastObject];
    }
    [newFeeds enumerateObjectsUsingBlock:^(NSDictionary *feedDict, NSUInteger idx, BOOL *stop) {
        Feed *feed = [self feedWithId:[feedDict objectForKey:@"id"]];
        feed.unreadCount = [feedDict objectForKey:@"unreadCount"];
        feed.folderId = [feedDict objectForKey:@"folderId"];
    }];
    
    for (NSNumber *feedId in feedsToDelete) {
        Feed *feed = [self feedWithId:feedId];
        [self deleteFeedOffline:feed]; //the feed will have been readded as new on server
    }
    [feedsToDelete removeAllObjects];
    
    for (NSString *urlString in feedsToAdd) {
        [self addFeedOffline:urlString];
    }
    [feedsToAdd removeAllObjects];
    
    //@{@"feedId": aFeedId, @"folderId": aFolderId}];
    for (NSDictionary *dict in feedsToMove) {
        [self moveFeedOfflineWithId:[dict objectForKey:@"feedId"] toFolderWithId:[dict objectForKey:@"folderId"]];
    }
    [feedsToMove removeAllObjects];
    
    [self.context processPendingChanges]; //Prevents crash if a feed has moved to another folder
    [self updateTotalUnreadCount];
}

- (Folder*)folderWithId:(NSNumber*)anId {
    [self.folderRequest setPredicate:[NSPredicate predicateWithFormat:@"myId == %@", anId]];
    NSArray *myFolders = [self.context executeFetchRequest:self.folderRequest error:nil];
    return (Folder*)[myFolders lastObject];
}

- (Feed*)feedWithId:(NSNumber*)anId {
    [self.feedRequest setPredicate:[NSPredicate predicateWithFormat:@"myId == %@", anId]];
    NSArray *myFeeds = [self.context executeFetchRequest:self.feedRequest error:nil];
    return (Feed*)[myFeeds lastObject];
}

- (Item*)itemWithId:(NSNumber *)anId {
    [self.itemRequest setPredicate:[NSPredicate predicateWithFormat:@"myId == %@", anId]];
    NSArray *myItems = [self.context executeFetchRequest:self.itemRequest error:nil];
    return (Item*)[myItems lastObject];
}

- (int)feedCount {
    [self.feedRequest setPredicate:nil];
    int count = [self.context countForFetchRequest:self.feedRequest error:nil];
    return count - 2;
}

- (int)itemCount {
    [self.itemRequest setPredicate:nil];
    int count = [self.context countForFetchRequest:self.itemRequest error:nil];
    return count;
}

- (void)updateFolderWithId:(NSNumber *)anId {
    NSNumber *lastMod = [self folderLastModified:anId];
    [self updateItemsWithLastModified:lastMod type:[NSNumber numberWithInt:UPDATE_FOLDER] andId:anId];
}

- (void)updateFeedWithId:(NSNumber*)anId {
    NSNumber *lastMod = [self feedLastModified:anId];
    if ([anId intValue] == -1) {
        [self updateItemsWithLastModified:lastMod type:[NSNumber numberWithInt:UPDATE_STARRED] andId:[NSNumber numberWithInt:0]];
    } else {
        [self updateItemsWithLastModified:lastMod type:[NSNumber numberWithInt:UPDATE_FEED] andId:anId];
    }
}

- (NSNumber*)folderLastModified:(NSNumber *)aFolderId {
    Folder *folder = [self folderWithId:aFolderId];
    NSNumber *lastFolderUpdate = folder.lastModified;
    NSNumber *lastSync = [NSNumber numberWithInt:[[NSUserDefaults standardUserDefaults] integerForKey:@"LastModified"]];
    return [NSNumber numberWithInt:MAX([lastFolderUpdate intValue], [lastSync intValue])];
}

- (NSNumber*)feedLastModified:(NSNumber *)aFeedId {
    Feed *feed = [self feedWithId:aFeedId];
    NSNumber *lastFeedUpdate = feed.extra.lastModified;
    NSNumber *lastSync = [NSNumber numberWithInt:[[NSUserDefaults standardUserDefaults] integerForKey:@"LastModified"]];
    return [NSNumber numberWithInt:MAX([lastFeedUpdate intValue], [lastSync intValue])];
}

- (void)updateItemsWithLastModified:(NSNumber*)lastMod type:(NSNumber*)aType andId:(NSNumber*)anId {
    NSDictionary *itemParams = @{@"lastModified": lastMod,
                                         @"type": aType,
                                           @"id": anId};
    
    [[OCAPIClient sharedClient] getPath:@"items/updated" parameters:itemParams success:^(AFHTTPRequestOperation *operation, id responseObject) {
        NSLog(@"URL: %@", operation.request.URL.absoluteString);
        NSDictionary *itemDict = (NSDictionary*)responseObject;
        //NSLog(@"New Items: %@", itemDict);
        NSArray *newItems = [NSArray arrayWithArray:[itemDict objectForKey:@"items"]];
        NSLog(@"New Item Count: %d", newItems.count);
        if (newItems.count > 0) {
            __block NSMutableSet *possibleDuplicateItems = [NSMutableSet new];
            [possibleDuplicateItems addObjectsFromArray:[newItems valueForKey:@"id"]];
            NSLog(@"Item count: %i; possibleDuplicateItems count: %i", newItems.count, possibleDuplicateItems.count);
            [self.itemRequest setPredicate:[NSPredicate predicateWithFormat: @"myId IN %@", possibleDuplicateItems]];
            
            [self.itemRequest setResultType:NSManagedObjectResultType];
            
            NSArray *duplicateItems = [self.context executeFetchRequest:self.itemRequest error:nil];
            NSLog(@"duplicateItems Count: %i", duplicateItems.count);
            
            for (NSManagedObject *item in duplicateItems) {
                NSLog(@"Deleting duplicate with title: %@", ((Item*)item).title);
                [self.context deleteObject:item];
            }
            
            __block NSMutableSet *feedsWithNewItems = [[NSMutableSet alloc] init];
            [newItems enumerateObjectsUsingBlock:^(NSDictionary *item, NSUInteger idx, BOOL *stop ) {
                NSNumber *myFeedId = [item objectForKey:@"feedId"];
                [feedsWithNewItems addObject:myFeedId];
                [self addItemFromDictionary:item];
            }];
            
            NSSortDescriptor *sort = [[NSSortDescriptor alloc] initWithKey:@"myId" ascending:NO];
            [self.itemRequest setSortDescriptors:[NSArray arrayWithObject:sort]];
            NSLog(@"Feeds with new items: %d", feedsWithNewItems.count);
            [feedsWithNewItems enumerateObjectsUsingBlock:^(NSNumber *feedId, BOOL *stop) {
                [self.itemRequest setPredicate:[NSPredicate predicateWithFormat: @"feedId == %@", feedId]];
                
                NSArray *feedItems = [self.context executeFetchRequest:self.itemRequest error:nil];
                NSMutableArray *filteredArray = [NSMutableArray arrayWithArray:[feedItems filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"unread == %@", [NSNumber numberWithBool:NO]]]];
                filteredArray = [NSMutableArray arrayWithArray:[filteredArray filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"starred == %@", [NSNumber numberWithBool:NO]]]];
                while (filteredArray.count > 500) {
                    Item *itemToRemove = [filteredArray lastObject];
                    NSLog(@"Deleting item with id %i and title %@", itemToRemove.myIdValue, itemToRemove.title);
                    [self.context deleteObject:itemToRemove];
                    [filteredArray removeLastObject];
                }
                
                NSArray *unreadItems = [feedItems filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"unread == %@", [NSNumber numberWithBool:YES]]];
                NSLog(@"Unread item count: %d", unreadItems.count);
                Feed *feed = [self feedWithId:feedId];
                feed.unreadCountValue = unreadItems.count;
            }];
        }
        
        switch ([aType intValue]) {
            case UPDATE_ALL: {
                NSLog(@"Finishing all item update");
                [self markItemsReadOffline:itemsToMarkRead];
                [itemsToMarkRead removeAllObjects];
                for (NSNumber *itemId in itemsToMarkUnread) {
                    [self markItemUnreadOffline:itemId];
                }
                [itemsToMarkUnread removeAllObjects];
                for (NSNumber *itemId in itemsToStar) {
                    [self starItemOffline:itemId];
                }
                [itemsToStar removeAllObjects];
                for (NSNumber *itemId in itemsToUnstar) {
                    [self unstarItemOffline:itemId];
                }
                [itemsToUnstar removeAllObjects];
                [self updateStarredCount];
                [self updateTotalUnreadCount];
                [[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:[[NSDate date] timeIntervalSince1970]] forKey:@"LastModified"];
            }
                break;
            case UPDATE_FOLDER: {
                NSLog(@"Finishing folder item update");
                Folder *folder = [self folderWithId:anId];
                folder.lastModified = [NSNumber numberWithInt:[[NSDate date] timeIntervalSince1970]];
            }
                break;
            case UPDATE_FEED:
            case UPDATE_STARRED: {
                NSLog(@"Finishing feed item update");
                Feed *feed = [self feedWithId:anId];
                feed.extra.lastModified = [NSNumber numberWithInt:[[NSDate date] timeIntervalSince1970]];
            }
                break;
                
            default:
                break;
        }
        
        [self updateStarredCount];
        [self updateTotalUnreadCount];
        [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkSuccess" object:self userInfo:nil];
        
    } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
        
        //feedsToUpdate;
        switch ([aType intValue]) {
            case UPDATE_ALL:
                NSLog(@"Doing single feed all item update");
                [self updateFeedItemsWithLastModified:lastMod type:aType andId:anId];
                break;
            case UPDATE_FOLDER: {
                NSLog(@"Doing single feed folder item update");
                //update feeds individually
                [self updateFeedItemsWithLastModified:lastMod type:aType andId:anId];
            }
                break;
            case UPDATE_FEED:
            case UPDATE_STARRED: {
                NSLog(@"Finishing feed item update");
                NSString *message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
                NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Updating Items", @"Title", message, @"Message", nil];
                [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
            }
                break;
                
            default:
                break;
        }
    }];
}


- (void)updateFeedItemsWithLastModified:(NSNumber*)lastMod type:(NSNumber*)aType andId:(NSNumber*)anId {
    __block NSMutableArray *operations = [NSMutableArray new];
    __block NSMutableArray *addedItems = [NSMutableArray new];
    __block OCAPIClient *client = [OCAPIClient sharedClient];
    
    //update feeds individually
    [self.feedRequest setPredicate:[NSPredicate predicateWithFormat:@"myId > 0"]];
    __block NSArray *allFeeds = [self.context executeFetchRequest:self.feedRequest error:nil];
    if ([aType intValue] == UPDATE_FOLDER) {
        allFeeds = [allFeeds filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"feedId == %@", anId]];
    }

    NSLog(@"Building operations");
    [allFeeds enumerateObjectsUsingBlock:^(Feed *feed, NSUInteger idx, BOOL *stop) {
        NSDictionary *itemParams = [NSDictionary dictionaryWithObjectsAndKeys:[self feedLastModified:feed.myId], @"lastModified",
                                    [NSNumber numberWithInt:0], @"type",
                                    feed.myId, @"id", nil];
        
        
        NSMutableURLRequest *itemURLRequest = [client requestWithMethod:@"GET" path:@"items/updated" parameters:itemParams];
        
        AFJSONRequestOperation *itemOperation = [AFJSONRequestOperation JSONRequestOperationWithRequest:itemURLRequest success:nil failure:nil];
        BOOL allowInvalid = [[NSUserDefaults standardUserDefaults] boolForKey:@"AllowInvalidSSLCertificate"];
        if (allowInvalid) {
            [itemOperation setWillSendRequestForAuthenticationChallengeBlock:^(NSURLConnection *connection, NSURLAuthenticationChallenge *challenge) {
                if ([challenge.protectionSpace.authenticationMethod isEqualToString:NSURLAuthenticationMethodServerTrust]) {
                    [challenge.sender useCredential:[NSURLCredential credentialForTrust:challenge.protectionSpace.serverTrust] forAuthenticationChallenge:challenge];
                }
            }];
        }
        NSLog(@"Adding operation %i", idx);
        [operations addObject:itemOperation];
    }];
    NSLog(@"Enqueing operations");
    [client enqueueBatchOfHTTPRequestOperations:operations progressBlock:^(NSUInteger numberOfFinishedOperations, NSUInteger totalNumberOfOperations) {
        NSLog(@"Feed %i of %i", numberOfFinishedOperations, totalNumberOfOperations);
    } completionBlock:^(NSArray *operations) {
        __block int errorCount = 0;
        __block NSMutableSet *feedsWithNewItems;
        [operations enumerateObjectsUsingBlock:^(AFJSONRequestOperation *itemOperation, NSUInteger idx, BOOL *stop) {
            //NSLog(@"New Items: %@", JSON);
            if (itemOperation.error) {
                ++errorCount;
            } else {
                NSDictionary *jsonDict = (NSDictionary *) itemOperation.responseJSON;
                NSArray *newItems = [NSArray arrayWithArray:[jsonDict objectForKey:@"items"]];
                //NSLog(@"New Item Count: %d", newItems.count);
                if (newItems.count > 0) {
                    [feedsWithNewItems addObject:[(NSDictionary*)[newItems objectAtIndex:0] objectForKey:@"feedId"]];
                    [addedItems addObjectsFromArray:newItems];
                    //NSLog(@"Feed: %@ (%d) adding %d for %d total items", feed.title, feed.unreadCountValue, newItems.count, addedItems.count);
                }
            }
        }];
        
        if (addedItems.count > 0) {
            __block NSMutableSet *possibleDuplicateItems = [NSMutableSet new];
            [possibleDuplicateItems addObjectsFromArray:[addedItems valueForKey:@"id"]];
            NSLog(@"Item count: %i; possibleDuplicateItems count: %i", addedItems.count, possibleDuplicateItems.count);
            [self.itemRequest setPredicate:[NSPredicate predicateWithFormat: @"myId IN %@", possibleDuplicateItems]];
            [self.itemRequest setResultType:NSManagedObjectResultType];
            NSError *error = nil;
            NSArray *duplicateItems = [self.context executeFetchRequest:self.itemRequest error:&error];
            NSLog(@"duplicateItems Count: %i", duplicateItems.count);
            
            for (NSManagedObject *item in duplicateItems) {
                //NSLog(@"Deleting duplicate with title: %@", ((Item*)item).title);
                [self.context deleteObject:item];
            }
            
            [addedItems enumerateObjectsUsingBlock:^(NSDictionary *item, NSUInteger idx, BOOL *stop ) {
                [self addItemFromDictionary:item];
            }];
            
            NSSortDescriptor *sort = [[NSSortDescriptor alloc] initWithKey:@"myId" ascending:NO];
            [self.itemRequest setSortDescriptors:[NSArray arrayWithObject:sort]];
            
            [feedsWithNewItems enumerateObjectsUsingBlock:^(NSNumber *feedId, BOOL *stop) {
                
                //[allFeeds enumerateObjectsUsingBlock:^(Feed *obj, NSUInteger idx, BOOL *stop) {
                
                [self.itemRequest setPredicate:[NSPredicate predicateWithFormat: @"feedId == %@", feedId]];
                
                NSError *error = nil;
                NSMutableArray *feedItems = [NSMutableArray arrayWithArray:[self.context executeFetchRequest:self.itemRequest error:&error]];
                
                while (feedItems.count > 500) {
                    Item *itemToRemove = [feedItems lastObject];
                    if (!itemToRemove.starredValue) {
                        if (!itemToRemove.unreadValue) {
                            NSLog(@"Deleting item with id %i and title %@", itemToRemove.myIdValue, itemToRemove.title);
                            [self.context deleteObject:itemToRemove];
                            [feedItems removeLastObject];
                        }
                    }
                }
            }];
            if ([aType intValue] == UPDATE_ALL) {
                
                [self markItemsReadOffline:itemsToMarkRead];
                [itemsToMarkRead removeAllObjects];
                for (NSNumber *itemId in itemsToMarkUnread) {
                    [self markItemUnreadOffline:itemId];
                }
                [itemsToMarkUnread removeAllObjects];
                for (NSNumber *itemId in itemsToStar) {
                    [self starItemOffline:itemId];
                }
                [itemsToStar removeAllObjects];
                for (NSNumber *itemId in itemsToUnstar) {
                    [self unstarItemOffline:itemId];
                }
                [itemsToUnstar removeAllObjects];
            }
            
        }
        [self updateStarredCount];
        [self updateTotalUnreadCount];
        if (errorCount > 0) {
            NSString *message = @"At least one feed failed to update properly. Try syncing again.";
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Updating Items", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        } else {
            [[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:[[NSDate date] timeIntervalSince1970]] forKey:@"LastModified"];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkSuccess" object:self userInfo:nil];
        }
    }];
}


- (void)updateItemsFirstTime {
    __block NSMutableArray *operations = [NSMutableArray new];
    __block NSMutableArray *addedItems = [NSMutableArray new];
    __block OCAPIClient *client = [OCAPIClient sharedClient];
    NSError *error = nil;
    [self.feedRequest setPredicate:nil];
    NSArray *feeds = [self.context executeFetchRequest:self.feedRequest error:&error];
    
    [feeds enumerateObjectsUsingBlock:^(Feed *feed, NSUInteger idx, BOOL *stop) {
        int batchSize = MAX(50, feed.unreadCountValue);
        NSDictionary *itemParams = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:batchSize], @"batchSize",
                                    [NSNumber numberWithInt:0], @"offset",
                                    [NSNumber numberWithInt:0], @"type",
                                    feed.myId, @"id",
                                    @"true", @"getRead", nil];
        
        NSMutableURLRequest *itemURLRequest = [client requestWithMethod:@"GET" path:@"items" parameters:itemParams];
        
        AFJSONRequestOperation *itemOperation = [AFJSONRequestOperation JSONRequestOperationWithRequest:itemURLRequest success:nil failure:nil];
        BOOL allowInvalid = [[NSUserDefaults standardUserDefaults] boolForKey:@"AllowInvalidSSLCertificate"];
        if (allowInvalid) {
            [itemOperation setWillSendRequestForAuthenticationChallengeBlock:^(NSURLConnection *connection, NSURLAuthenticationChallenge *challenge) {
                if ([challenge.protectionSpace.authenticationMethod isEqualToString:NSURLAuthenticationMethodServerTrust]) {
                    [challenge.sender useCredential:[NSURLCredential credentialForTrust:challenge.protectionSpace.serverTrust] forAuthenticationChallenge:challenge];
                }
                
            }];
        }
        
        [operations addObject:itemOperation];
    }];
    
    [client enqueueBatchOfHTTPRequestOperations:operations progressBlock:^(NSUInteger numberOfFinishedOperations, NSUInteger totalNumberOfOperations) {
        NSLog(@"Feed %i of %i", numberOfFinishedOperations, totalNumberOfOperations);
    } completionBlock:^(NSArray *operations) {
        __block int errorCount = 0;
        [operations enumerateObjectsUsingBlock:^(AFJSONRequestOperation *itemOperation, NSUInteger idx, BOOL *stop) {
            //NSLog(@"New Items: %@", JSON);
            if (itemOperation.error) {
                ++errorCount;
            } else {
                //NSLog(@"New Items: %@", JSON);
                NSDictionary *jsonDict = (NSDictionary *) itemOperation.responseJSON;
                NSArray *newItems = [NSArray arrayWithArray:[jsonDict objectForKey:@"items"]];
                [addedItems addObjectsFromArray:newItems];
            }
        }];
        [addedItems enumerateObjectsUsingBlock:^(NSDictionary *item, NSUInteger idx, BOOL *stop ) {
            [self addItemFromDictionary:item];
        }];
        [self updateStarredCount];
        [self updateTotalUnreadCount];
        if (errorCount > 0) {
            NSString *message = @"At least one feed failed to update properly. Try syncing again.";
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Updating Items", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        } else {
            [[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:[[NSDate date] timeIntervalSince1970]] forKey:@"LastModified"];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkSuccess" object:self userInfo:nil];
        }
    }];
}

- (void)updateReadItems:(NSArray *)items {
    [self.itemRequest setPredicate:[NSPredicate predicateWithFormat: @"myId IN %@", items]];

    NSError *error = nil;
    NSArray *allItems = [self.context executeFetchRequest:self.itemRequest error:&error];
    if (!allItems || error) {
        NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
        abort();
    }
    NSLog(@"Count: %i", items.count);
    
    [allItems enumerateObjectsUsingBlock:^(Item *item, NSUInteger idx, BOOL *stop) {
        Feed *feed = [self feedWithId:item.feedId];
        if (item.unreadValue) {
            ++feed.unreadCountValue;
        } else {
            --feed.unreadCountValue;
        }
    }];
    [self updateTotalUnreadCount];
}

- (void)updateFolderUnreadCount {
    NSArray *folders = [self.context executeFetchRequest:self.folderRequest error:nil];
    [folders enumerateObjectsUsingBlock:^(Folder *folder, NSUInteger idx, BOOL *stop) {
        self.feedRequest.predicate = [NSPredicate predicateWithFormat:@"folderId == %@", folder.myId];
        NSArray *feeds = [self.context executeFetchRequest:self.feedRequest error:nil];
        folder.unreadCountValue = [[feeds valueForKeyPath:@"@sum.unreadCount"] integerValue];
    }];
}

- (void)updateTotalUnreadCount {
    [self.feedRequest setPredicate:[NSPredicate predicateWithFormat:@"myId > 0"]];
    NSArray *feeds = [self.context executeFetchRequest:self.feedRequest error:nil];
    int totalUnreadCount = [[feeds valueForKeyPath:@"@sum.unreadCount"] integerValue];
    [self feedWithId:[NSNumber numberWithInt:-2]].unreadCountValue = totalUnreadCount;
    [UIApplication sharedApplication].applicationIconBadgeNumber = totalUnreadCount;
    [self updateFolderUnreadCount];
    [self saveContext];
}

- (void)updateStarredCount {
    [self.itemRequest setPredicate:[NSPredicate predicateWithFormat: @"starred == 1"]];
    
    NSError *error = nil;
    NSArray *starredItems = [self.context executeFetchRequest:self.itemRequest error:&error];
    if (!starredItems || error) {
        NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
        abort();
    }
    NSLog(@"Starred Count: %i", starredItems.count);
    
    error = nil;
    NSArray *feeds = [self.context executeFetchRequest:self.feedsRequest error:&error];
    if (!feeds || error) {
        NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
        abort();
    }

    Feeds *theFeeds = [feeds lastObject];
    theFeeds.starredCountValue = starredItems.count;
    
    [[self feedWithId:[NSNumber numberWithInt:-1]] setUnreadCountValue:starredItems.count];
    [self saveContext];
}

- (void)addFolderOffline:(NSString*)name {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        NSDictionary *params = @{@"name": name};
        
        [[OCAPIClient sharedClient] postPath:@"folders" parameters:params success:^(AFHTTPRequestOperation *operation, id responseObject) {
            //NSLog(@"Folders: %@", responseObject);
            __unused int newFolderId = [self addFolder:responseObject];
        
        } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
            NSString *message;
            switch (operation.response.statusCode) {
                case 409:
                    message = @"The folder already exists.";
                    break;
                case 422:
                    message = @"The folder name is invalid.";
                    break;
                default:
                    message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'.", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
                    break;
            }
            
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Adding Folder", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        }];
        
    } else {
        //offline
        [foldersToAdd addObject:name];
        Folder *newFolder = [NSEntityDescription insertNewObjectForEntityForName:@"Folder" inManagedObjectContext:self.context];
        newFolder.myId = [NSNumber numberWithInt:10000 + foldersToAdd.count];
        newFolder.name = name;
    }
    [self updateTotalUnreadCount];
}

- (void)deleteFolderOffline:(Folder*)folder {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        NSString *path = [NSString stringWithFormat:@"folders/%@", [folder.myId stringValue]];
        [[OCAPIClient sharedClient] deletePath:path parameters:nil success:^(AFHTTPRequestOperation *operation, id responseObject) {
            NSLog(@"Success");
        } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
            NSLog(@"Failure");
            NSString *message;
            switch (operation.response.statusCode) {
                case 404:
                    message = @"The folder does not exist.";
                    break;
                default:
                    message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'.", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
                    break;
            }
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Deleting Folder", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        }];
    } else {
        //offline
        [foldersToDelete addObject:folder.myId];
    }
    [self deleteFolder:folder];
}

- (void)renameFolderOfflineWithId:(NSNumber*)anId To:(NSString*)newName {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        NSDictionary *params = @{@"name": newName};
        NSString *path = [NSString stringWithFormat:@"folders/%@", [anId stringValue]];
        [[OCAPIClient sharedClient] putPath:path parameters:params success:^(AFHTTPRequestOperation *operation, id responseObject) {
            NSLog(@"Success");
        } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
            NSLog(@"Failure");
            NSString *message;
            switch (operation.response.statusCode) {
                case 404:
                    message = @"The folder does not exist.";
                    break;
                case 409:
                    message = @"A folder with this name already exists.";
                    break;
                case 422:
                    message = @"The folder name is invalid";
                    break;
                default:
                    message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'.", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
                    break;
            }
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Moving Feed", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        }];
    } else {
        //offline
        [foldersToRename addObject:@{@"folderId": anId, @"name": newName}];
    }
    [[self folderWithId:anId] setName:newName];
}

- (void)addFeedOffline:(NSString *)urlString {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        NSDictionary *params = [NSDictionary dictionaryWithObjectsAndKeys:urlString, @"url", [NSNumber numberWithInt:0], @"folderId", nil];
        
        [[OCAPIClient sharedClient] postPath:@"feeds" parameters:params success:^(AFHTTPRequestOperation *operation, id responseObject) {
            //NSLog(@"Feeds: %@", responseObject);
            
            int newFeedId = [self addFeed:responseObject];
            NSDictionary *params = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithInt:200], @"batchSize",
                                    [NSNumber numberWithInt:0], @"offset",
                                    [NSNumber numberWithInt:0], @"type",
                                    [NSNumber numberWithInt:newFeedId], @"id",
                                    [NSNumber numberWithInt:1], @"getRead", nil];
            
            [[OCAPIClient sharedClient] getPath:@"items" parameters:params success:^(AFHTTPRequestOperation *operation, id responseObject) {
                NSDictionary *jsonDict = (NSDictionary *) responseObject;
                NSArray *newItems = [NSArray arrayWithArray:[jsonDict objectForKey:@"items"]];
                [newItems enumerateObjectsUsingBlock:^(NSDictionary *item, NSUInteger idx, BOOL *stop ) {
                    [self addItemFromDictionary:item];
                }];
                
            } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
                NSString *message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
                NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Retrieving Items", @"Title", message, @"Message", nil];
                [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
            }];
        } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
            NSString *message;
            switch (operation.response.statusCode) {
                case 409:
                    message = @"The feed already exists";
                    break;
                case 422:
                    message = @"The feed could not be read. It most likely contains errors";
                    break;
                default:
                    message = [NSString stringWithFormat:@"The server repsonded '%@' and the error reported was '%@'", [NSHTTPURLResponse localizedStringForStatusCode:operation.response.statusCode], [error localizedDescription]];
                    break;
            }
            
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Adding Feed", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        }];
    
    } else {
        //offline
        [feedsToAdd addObject:urlString];
        Feed *newFeed = [NSEntityDescription insertNewObjectForEntityForName:@"Feed" inManagedObjectContext:self.context];
        newFeed.myId = [NSNumber numberWithInt:10000 + feedsToAdd.count];
        newFeed.url = urlString;
        newFeed.title = urlString;
        newFeed.faviconLink = @"favicon";
        newFeed.added = [NSNumber numberWithInt:1];
        newFeed.folderId = [NSNumber numberWithInt:0];
        newFeed.unreadCount = [NSNumber numberWithInt:0];
        newFeed.link = @"";
        //[feedsToDelete addObject:[NSNumber numberWithInt:10000 + feedsToAdd.count]]; //should be deleted when we get the real feed
    }
    [self updateTotalUnreadCount];
}

- (void) deleteFeedOffline:(Feed*)feed {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        NSString *path = [NSString stringWithFormat:@"feeds/%@", [feed.myId stringValue]];
        [[OCAPIClient sharedClient] deletePath:path parameters:nil success:^(AFHTTPRequestOperation *operation, id responseObject) {
            NSLog(@"Success");
        } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
            NSLog(@"Failure");
            NSString *message = [NSString stringWithFormat:@"The error reported was '%@'", [error localizedDescription]];
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Deleting Feed", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        }];
    } else {
        //offline
        [feedsToDelete addObject:feed.myId];
    }
    [self deleteFeed:feed];
}

- (void)moveFeedOfflineWithId:(NSNumber *)aFeedId toFolderWithId:(NSNumber *)aFolderId {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        NSDictionary *params = @{@"folderId": aFolderId};
        NSString *path = [NSString stringWithFormat:@"feeds/%@/move", [aFeedId stringValue]];
        [[OCAPIClient sharedClient] putPath:path parameters:params success:^(AFHTTPRequestOperation *operation, id responseObject) {
            NSLog(@"Success");
        } failure:^(AFHTTPRequestOperation *operation, NSError *error) {
            NSLog(@"Failure");
            NSString *message = [NSString stringWithFormat:@"The error reported was '%@'", [error localizedDescription]];
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:@"Error Moving Feed", @"Title", message, @"Message", nil];
            [[NSNotificationCenter defaultCenter] postNotificationName:@"NetworkError" object:self userInfo:userInfo];
        }];
    } else {
        //offline
        [feedsToMove addObject:@{@"feedId": aFeedId, @"folderId": aFolderId}];
    }
}

- (void)markItemsReadOffline:(NSArray *)itemIds {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        [[OCAPIClient sharedClient] putPath:@"items/read/multiple" parameters:[NSDictionary dictionaryWithObject:itemIds forKey:@"items"] success:nil failure:nil];
    } else {
        //offline
        for (NSNumber *itemId in itemIds) {
            NSInteger i = [itemsToMarkUnread indexOfObject:itemId];
            if (i != NSNotFound) {
                [itemsToMarkUnread removeObject:itemId];
            }
        }
        [itemsToMarkRead addObjectsFromArray:itemIds];
    }
    [self updateReadItems:itemIds];
}

- (void)markItemUnreadOffline:(NSNumber*)itemId {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        [[OCAPIClient sharedClient] putPath:@"items/unread/multiple" parameters:[NSDictionary dictionaryWithObject:itemId forKey:@"items"] success:nil failure:nil];
    } else {
        //offline
        NSInteger i = [itemsToMarkRead indexOfObject:itemId];
        if (i != NSNotFound) {
            [itemsToMarkRead removeObject:itemId];
        }
        [itemsToMarkUnread addObject:itemId];
    }
    [self updateReadItems:@[itemId]];
}

- (void)starItemOffline:(NSNumber*)itemId {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        Item *item = [self itemWithId:itemId];
        if (item) {
            NSString *path = [NSString stringWithFormat:@"items/%@/%@/star", [item.feedId stringValue], item.guidHash];
            [[OCAPIClient sharedClient] putPath:path parameters:nil success:nil failure:nil];
        }
    } else {
        //offline
        NSInteger i = [itemsToUnstar indexOfObject:itemId];
        if (i != NSNotFound) {
            [itemsToUnstar removeObject:itemId];
        }
        [itemsToStar addObject:itemId];
    }
    [self updateStarredCount];
}

- (void)unstarItemOffline:(NSNumber*)itemId {
    if ([[OCAPIClient sharedClient] networkReachabilityStatus] > 0) {
        //online
        Item *item = [self itemWithId:itemId];
        if (item) {
            NSString *path = [NSString stringWithFormat:@"items/%@/%@/unstar", [item.feedId stringValue], item.guidHash];
            [[OCAPIClient sharedClient] putPath:path parameters:nil success:nil failure:nil];
        }
    } else {
        //offline
        NSInteger i = [itemsToStar indexOfObject:itemId];
        if (i != NSNotFound) {
            [itemsToStar removeObject:itemId];
        }
        [itemsToUnstar addObject:itemId];
    }
    [self updateStarredCount];
}

- (NSFetchRequest *)feedsRequest {
    if (!feedsRequest) {
        feedsRequest = [[NSFetchRequest alloc] init];
        [feedsRequest setEntity:[NSEntityDescription entityForName:@"Feeds" inManagedObjectContext:self.context]];
    }
    return feedsRequest;
}

- (NSFetchRequest *)feedRequest {
    if (!feedRequest) {
        feedRequest = [[NSFetchRequest alloc] init];
        [feedRequest setEntity:[NSEntityDescription entityForName:@"Feed" inManagedObjectContext:self.context]];
        feedRequest.predicate = nil;
    }
    return feedRequest;
}

- (NSFetchRequest *)folderRequest {
    if (!folderRequest) {
        folderRequest = [[NSFetchRequest alloc] init];
        [folderRequest setEntity:[NSEntityDescription entityForName:@"Folder" inManagedObjectContext:self.context]];
        folderRequest.predicate = nil;
    }
    return folderRequest;
}

- (NSFetchRequest *)itemRequest {
    if (!itemRequest) {
        itemRequest = [[NSFetchRequest alloc] init];
        [itemRequest setEntity:[NSEntityDescription entityForName:@"Item" inManagedObjectContext:self.context]];
        itemRequest.predicate = nil;
    }
    return itemRequest;
}

- (NSURL*) documentsDirectoryURL {
    return [[[NSFileManager defaultManager] URLsForDirectory:NSDocumentDirectory inDomains:NSUserDomainMask] lastObject];
}

@end
